/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.web.tomcat.service.session;

import java.util.HashMap;
import java.util.Map;

import org.jboss.web.tomcat.service.session.distributedcache.spi.DistributedCacheManager;

/**
 * Implementation of a clustered session where the replication granularity
 * level is attribute based; that is, we replicate only the dirty attributes.
 * <p/>
 * Note that the isolation level of the cache dictates the
 * concurrency behavior. Also note that session and its associated attribtues are stored in different nodes.
 * This will be ok since cache will take care of concurrency. When replicating, we will need to replicate both
 * session and its attributes.</p>
 *
 * @author Ben Wang
 * @author Brian Stansberry
 * 
 * @version $Revision: 56542 $
 */
class AttributeBasedClusteredSession
   extends ClusteredSession
{
   static final long serialVersionUID = -5625209785550936713L;
   /**
    * Descriptive information describing this Session implementation.
    */
   protected static final String info = "AttributeBasedClusteredSession/1.0";

   // Transient map to store attr changes for replication.
   private transient Map<String, Object> attrModifiedMap_ = new HashMap<String, Object>();
   // Note that the removed attr is intentionally stored in a map 
   // instead of a Set so it is faster to lookup and remove.
   private transient Map<String, Object> attrRemovedMap_ = new HashMap<String, Object>();

   // ------------------------------------------------------------ Constructors
   

   public AttributeBasedClusteredSession(ClusteredManager manager)
   {
      super(manager);
   }

   
   // ----------------------------------------------- Overridden Public Methods

   @Override
   public String getInfo()
   {
      return (info);
   }

   /**
    * Override the superclass to additionally reset this class' fields.
    * <p>
    * <strong>NOTE:</strong> It is not anticipated that this method will be
    * called on a ClusteredSession, but we are overriding the method to be
    * thorough.
    * </p>
    */
   @Override
   public void recycle()
   {
      super.recycle();

      clearAttrChangedMaps();
   }

   // -------------------------------------------- Overridden Protected Methods

//   /**
//    * Populate the attributes stored in the distributed store to local transient ones.
//    */
//   @Override
//   protected void populateAttributes(Map<String, Object> distributedCacheAttributes)
//   {
//      Map<String, Object> existing = getAttributesInternal();
//      Map<String, Object> excluded = removeExcludedAttributes(existing);
//      existing.clear();
//      
//      existing.putAll(distributedCacheAttributes);
//      if (excluded != null)
//         existing.putAll(excluded);
//      
//      attrModifiedMap_.clear();
//      attrRemovedMap_.clear();
//   }

   @Override
   protected Object getAttributeInternal(String name)
   {
      Object result = getAttributesInternal().get(name);

      // Do dirty check even if result is null, as w/ SET_AND_GET null
      // still makes us dirty (ensures timely replication w/o using ACCESS)
      if (isGetDirty(result) && !replicationExcludes.contains(name))
      {
         attributeChanged(name, result, false);
      }

      return result;
   }

   @Override
   protected Object removeAttributeInternal(String name, 
                                            boolean localCall, 
                                            boolean localOnly)
   {
      Object result = getAttributesInternal().remove(name);
      if (localCall && !replicationExcludes.contains(name))
         attributeChanged(name, result, true);
      return result;
   }

   @Override
   protected Object setAttributeInternal(String key, Object value)
   {
      Object old = getAttributesInternal().put(key, value);
      if (!replicationExcludes.contains(key))
         attributeChanged(key, value, false);
      return old;
   }

   /**
    * Overrides the superclass version to read in the attributes.
    */
   @Override
   protected synchronized void replicateAttributes()
   {
      // Go thru the attribute change list
      String myRealId = getRealId();
      if (isSessionAttributeMapDirty())
      {
         DistributedCacheManager distributedCacheManager = getDistributedCacheManager();
         
         // Go thru the modified attr list first
         int modCount = attrModifiedMap_.size();
         if (modCount == 1)
         {
            for (Map.Entry<String, Object> entry : attrModifiedMap_.entrySet())
            {
               distributedCacheManager.putAttribute(myRealId, entry.getKey(), entry.getValue());
            }
         }
         else if (modCount > 0)
         {
            // It's more efficient to write a map than 2 method calls,
            // plus it reduces the number of CacheListener notifications
            distributedCacheManager.putAttribute(myRealId, attrModifiedMap_);
         }
         
         // Go thru the remove attr list
         if (attrRemovedMap_.size() > 0)
         {         
            for (String key : attrRemovedMap_.keySet())
            {
               distributedCacheManager.removeAttribute(myRealId, key);
            }
         }
         
         clearAttrChangedMaps();
      }
   }

   // ------------------------------------------------------- Private Methods

   private synchronized void attributeChanged(String key, Object value, boolean removal)
   {
      if (removal)
      {
         if (attrModifiedMap_.containsKey(key))
         {
            attrModifiedMap_.remove(key);
         }
         attrRemovedMap_.put(key, value);
      }
      else
      {
         if (attrRemovedMap_.containsKey(key))
         {
            attrRemovedMap_.remove(key);
         }
         attrModifiedMap_.put(key, value);
      }
      sessionAttributesDirty();
   }

   private synchronized void clearAttrChangedMaps()
   {
      attrRemovedMap_.clear();
      attrModifiedMap_.clear();
   }
}
